"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = __importDefault(require("util"));
const os_1 = __importDefault(require("os"));
const express_1 = __importDefault(require("express"));
const backtrace_service_1 = require("backtrace-service");
const packageJson = require('../package.json');
const candidates_1 = require("./application/candidates");
const utilities_1 = require("./application/utilities");
const errors_1 = require("./application/errors");
const workerPool_1 = require("./application/workerPool");
const log = backtrace_service_1.Logger.getLogger();
/**
 * Create Backtrace Service
 */
const endpoint = '/api/similarity';
const service = backtrace_service_1.BacktraceService.bootstrap('similarity', 6454, {}, express_1.default());
/**
 * Set up defaults
 */
process.setMaxListeners(0);
const clusterSupport = service.configuration.clusterSupport;
const requestWorkers = service.configuration.requestWorkers;
const requestTaskLimit = service.configuration.requestTaskLimit;
const globalTaskLimitSleepMs = service.configuration.globalTaskLimitSleepMs;
const globalTaskLimit = service.configuration.globalTaskLimit;
const requiredPositiveInts = {
    clusterSupport,
    requestWorkers,
    globalTaskLimit,
    globalTaskLimitSleepMs,
    requestTaskLimit,
};
Object.keys(requiredPositiveInts).forEach((varname) => {
    const value = requiredPositiveInts[varname];
    if (!Number.isInteger(value) || value <= 0) {
        throw new Error(`must provide a positive integer for ${varname}`);
    }
});
log.info(`
==Cluster support== 
  cpus: ${os_1.default.cpus().length}
  clusterSupport: ${clusterSupport} 
  requestWorkers: ${requestWorkers}`);
log.info(`
==Worker Limits== 
  globalTaskLimit: ${globalTaskLimit}
  globalTaskLimitSleepMs: ${globalTaskLimitSleepMs} 
  requestTaskLimit: ${requestTaskLimit}`);
/**
 * Set up Worker Pools
 */
const poolDistance = new workerPool_1.WorkerPool(requestWorkers);
poolDistance.setMaxListeners(0);
const poolSummary = new workerPool_1.WorkerPool(requestWorkers);
poolSummary.setMaxListeners(0);
/**
 * Set up semaphore to limit global tasks to protect server from OOM
 */
const globalLimit = utilities_1.semaphore(globalTaskLimit, globalTaskLimitSleepMs);
/**
 * Get candidates for a given fingerprint and requested type of similarity processing
 */
const candidatesRequiredFields = [
    'project',
    'fingerprint',
    'candidacy',
    'limit',
];
const distanceVariablesRequiredNumbers = [
    'threshold',
    'truncate',
    'intersection',
    'distance',
];
service.app.post(`${endpoint}/candidates`, service.identityManager.authenticateRequest(), async (req, res, _next) => {
    var _a, _b, _c;
    let requestState = {
        cancelRequest: false,
    };
    req.on('close', () => {
        requestState.cancelRequest = true;
    });
    const authRequest = req;
    const missing = errors_1.getMissingProperties(req.body, candidatesRequiredFields);
    if (missing) {
        service.identityManager.bypassProjectAccessCheck(res);
        return errors_1.sendClientError(res, errors_1.missingFields(missing));
    }
    const { project, fingerprint, candidacy, limit } = req.body;
    if (!service.identityManager.checkProjectAccess(req, res, [project])) {
        return res.send();
    }
    // set up variable defaults and processing function based on type of similarity,
    // for each similarity request
    let clientError = null;
    const similarityRequests = [];
    for (let i = 0; !clientError && i < candidacy.length; i++) {
        const c = candidacy[i];
        switch (c.type) {
            case 'distance': {
                distanceVariablesRequiredNumbers.forEach((k) => {
                    if (!Number.isInteger(c[k])) {
                        clientError = errors_1.errors.malformedRequest;
                    }
                });
                similarityRequests[i] = {
                    type: c.type,
                    variables: {
                        truncate: c.truncate,
                        distance: c.distance,
                        intersection: c.intersection,
                        threshold: c.threshold,
                    },
                };
                break;
            }
            default: {
                clientError = errors_1.noSuchOption(c.type);
            }
        }
    }
    if (clientError) {
        return errors_1.sendClientError(res, clientError);
    }
    const crUrl = (_a = authRequest.coronerAuth) === null || _a === void 0 ? void 0 : _a.url;
    const cfg = (_b = authRequest.coronerAuth) === null || _b === void 0 ? void 0 : _b.cfg;
    const token = (_c = authRequest.coronerAuth) === null || _c === void 0 ? void 0 : _c.token;
    // Middleware handles if unauthed so handling this as internal error.
    if (!crUrl || !cfg || !token) {
        return errors_1.sendInternalError(res, new Error('missing url, cfg, or token.'));
    }
    const queryUrl = `${utilities_1.getUrlForEndpoint(crUrl, cfg, 'api/query')}`;
    const url = `${queryUrl}?project=${project}&token=${token}`;
    let mainFpData;
    try {
        const results = await candidates_1.getDataForFingerprint(fingerprint, url);
        mainFpData = utilities_1.refineFingerprint(results, fingerprint);
    }
    catch (_d) {
        return errors_1.sendClientError(res, errors_1.errors.noSuchObject);
    }
    const candidates = await candidates_1.getDataForFingerprintsByTimestampAtLeast(mainFpData.first_seen, url, limit);
    const candidateFingerprints = Object.keys(candidates);
    // Multiple workers handle comparisons, promises keep track of when they all finish.
    const similarityResults = [];
    const taskPromiseFns = [];
    // For each candidate fingerprint, run each similarity request
    candidateFingerprints.forEach((fp) => {
        const candidateRefined = utilities_1.refineFingerprint(candidates, fp);
        similarityRequests.forEach((simReq, index) => {
            // Initialize the response object. Some fields will be updated by workers below.
            similarityResults[index] = {
                fingerprint: fingerprint,
                callstack: mainFpData.callstack,
                meta: {
                    disqualified: {
                        byThreshold: 0,
                        byIntersection: 0,
                        byDistance: 0,
                    },
                    qualified: 0,
                },
                candidates: [],
            };
            //  Create a promise to run the task and resolve/reject when it finishes
            const taskPromiseFn = () => {
                return new Promise((resolve, reject) => {
                    poolDistance.runTask({
                        type: simReq.type,
                        candidate: candidateRefined,
                        variables: simReq.variables,
                        mainFingerprint: mainFpData,
                    }, (err, result) => {
                        const r = result;
                        if (err) {
                            log.error(err);
                            return reject(err.message);
                        }
                        if (r.candidate) {
                            similarityResults[index].candidates.push(r.candidate);
                        }
                        similarityResults[index].meta.disqualified.byIntersection +=
                            r.byIntersection;
                        similarityResults[index].meta.disqualified.byDistance +=
                            r.byDistance;
                        similarityResults[index].meta.disqualified.byThreshold +=
                            r.byThreshold;
                        return resolve();
                    });
                });
            };
            taskPromiseFns.push(taskPromiseFn);
        });
    });
    await utilities_1.limitTasks(taskPromiseFns, requestTaskLimit, globalLimit, requestState)
        .then(function (_resps) {
        if (requestState.cancelRequest) {
            return;
        }
        similarityResults.forEach((similarityResult) => {
            const candidates = similarityResult.candidates;
            candidates.sort(utilities_1.compareCandidates);
            similarityResult.meta.qualified = similarityResult.candidates.length;
        });
        return res.send({
            status: 'ok',
            results: similarityResults,
        });
    })
        .catch(function (err) {
        const compactError = new Error(err.message);
        compactError.stack = err.stack;
        log.error(compactError);
        return errors_1.sendClientError(res, errors_1.errors.internalError);
    });
});
/**
 *  Summary information of similarity across fingerprints in an entire project
 */
const summaryRequiredFields = ['project', 'filter', 'limit', 'candidacy'];
service.app.post(`${endpoint}/summary`, service.identityManager.authenticateRequest(), async (req, res, _next) => {
    var _a, _b, _c;
    let requestState = {
        cancelRequest: false,
    };
    req.on('close', () => {
        requestState.cancelRequest = true;
    });
    const authRequest = req;
    const missing = errors_1.getMissingProperties(req.body, summaryRequiredFields);
    if (missing) {
        service.identityManager.bypassProjectAccessCheck(res);
        return errors_1.sendClientError(res, errors_1.missingFields(missing));
    }
    const { project, filter, limit, candidacy } = req.body;
    if (!service.identityManager.checkProjectAccess(req, res, [project])) {
        return res.send();
    }
    // set up variable defaults based on type of similarity
    let similarityRequest;
    switch (candidacy.type) {
        case 'distance': {
            distanceVariablesRequiredNumbers.forEach((k) => {
                if (!Number.isInteger(candidacy[k])) {
                    return errors_1.sendClientError(res, errors_1.errors.malformedRequest);
                }
            });
            similarityRequest = {
                type: candidacy.type,
                variables: {
                    truncate: candidacy.truncate,
                    distance: candidacy.distance,
                    intersection: candidacy.intersection,
                    threshold: candidacy.threshold,
                },
            };
            break;
        }
        default: {
            return errors_1.sendClientError(res, errors_1.noSuchOption(candidacy.type));
        }
    }
    const crUrl = (_a = authRequest.coronerAuth) === null || _a === void 0 ? void 0 : _a.url;
    const cfg = (_b = authRequest.coronerAuth) === null || _b === void 0 ? void 0 : _b.cfg;
    const token = (_c = authRequest.coronerAuth) === null || _c === void 0 ? void 0 : _c.token;
    // Middleware handles if unauthed so handling this as internal error.
    if (!crUrl || !cfg || !token) {
        return errors_1.sendInternalError(res, new Error('missing url, cfg, or token.'));
    }
    const queryUrl = `${utilities_1.getUrlForEndpoint(crUrl, cfg, 'api/query')}`;
    const url = `${queryUrl}?project=${project}&token=${token}`;
    let resultsRefined;
    try {
        const results = await candidates_1.getDataForFingerprintsByQuery(url, filter, limit);
        const resultFps = Object.keys(results);
        resultsRefined = resultFps.map((fp) => {
            return utilities_1.refineFingerprint(results, fp);
        });
    }
    catch (err) {
        log.debug(`Possibly malformed request. err: ${util_1.default.inspect(err, {
            depth: 5,
        })}. filter: ${util_1.default.inspect(filter, { depth: 5 })}`);
        return errors_1.sendClientError(res, errors_1.errors.malformedRequest);
    }
    const taskPromiseFns = [];
    const summaryResults = [];
    // For each fingerprint, run each similarity request
    resultsRefined.forEach((data) => {
        //  Create a promise to run the task and resolve/reject when it finishes
        const taskPromiseFn = () => {
            return new Promise((resolve, reject) => {
                poolSummary.runTask({
                    type: `${similarityRequest.type}-summary`,
                    candidates: resultsRefined,
                    variables: similarityRequest.variables,
                    mainFingerprint: data,
                }, (err, result) => {
                    const r = result;
                    if (err) {
                        log.error(err);
                        return reject(err.message);
                    }
                    summaryResults.push(r);
                    return resolve();
                });
            });
        };
        taskPromiseFns.push(taskPromiseFn);
    });
    await utilities_1.limitTasks(taskPromiseFns, requestTaskLimit, globalLimit, requestState)
        .then(function (_resps) {
        if (requestState.cancelRequest) {
            return;
        }
        summaryResults.sort(utilities_1.compareSummary);
        return res.send({
            status: 'ok',
            results: summaryResults,
        });
    })
        .catch(function (err) {
        const compactError = new Error(err.message);
        compactError.stack = err.stack;
        log.error(compactError);
        return errors_1.sendClientError(res, errors_1.errors.internalError);
    });
});
/**
 * Service Version
 */
service.app.get(`${endpoint}/version`, (_req, res) => {
    service.identityManager.bypassProjectAccessCheck(res);
    res.json({
        response: { status: 'ok' },
        _: {
            version: packageJson.version,
        },
    });
});
/**
 * Start
 */
service.start(clusterSupport);
exports.default = service;
//# sourceMappingURL=main.js.map